#include <math.h>

#include "polynomial.h"

int poly_load(FILE *fin, poly_t F) {
    int i, deg;

    fscanf(fin, "%d", &deg);
    F->deg = deg;

    for (i = 0; i <= deg; ++i) {
        mpz_init(F->k[i]);
        gmp_fscanf(fin, "%Zd", F->k[i]);
        F->d[i] = mpz_get_d(F->k[i]);
    }
    
    fscanf(fin, "%d", &deg);
    F->realRootsNumber = deg;
    for (i = 0; i < deg; ++i) {
        fscanf(fin, "%lf", &F->roots[i]);
    }
    
    mpz_init(F->ki);
    mpz_init(F->yi);
    
    return 0;
}

void poly_destroy(poly_t F)
{
    int i;
    
    for (i = 0; i <= F->deg; ++i) {
        mpz_clear(F->k[i]);
    }
    
    mpz_clear(F->yi);
    mpz_clear(F->ki);
}

int poly_has_roots_in(poly_t F, int64_t x, int64_t y) {
    int i;
    double xy = (double)x/(double)y;
    
    for(i=0; i < F->realRootsNumber && F->roots[i] < xy; ++i);

    if (i < F->realRootsNumber && (F->roots[i] * y <= x + BLOCK_SIZE))
        return 1;

    return 0;
}

void poly_value(poly_t F, int64_t x, int64_t y, mpz_t val) {
    int i;
    
    mpz_set_ui(val,0);
    mpz_set_si(F->yi, 1);
    
    for (i = F->deg; i >= 0; --i) {
        mpz_mul(F->ki, F->k[i], F->yi);
        mpz_mul_si(F->yi, F->yi, y);
        
        mpz_mul_si(val, val, x);
        mpz_add(val, val, F->ki);
    }
}

/*double estimatePoly(poly *F, int64_t x, int64_t y) {
    double yx = (double)y / (double)x;
    double xd = pow((double)x, F->deg);
    double res = 0.0;
    int i;
    for (i = F->deg; i >= 0; --i) {
        res += xd * F->d[i];
        xd *= yx;
    }
    return log(fabs(res)) * (LOG_SCALE / log(2.0));
}*/

double poly_estimate(poly_t F, int64_t x, int64_t y) {
    int i;
    double res = 0.0;
    double xy = (double)x / (double)y;
    double yd = pow((double)y, F->deg);

    for (i = 0; i <= F->deg; ++i) {
        res += yd * F->d[i];
        yd *= xy;
    }
    return log(fabs(res)) * (LOG_SCALE / log(2.0));
}


void poly_lift_root(poly_t F, int64_t invdf, int64_t p, int64_t* pk, int64_t* r) {
    mpz_t tval;
    
    mpz_init(tval); 
    poly_value(F, *r, 1, tval);

    mpz_mul_si(tval, tval, -invdf);

    if (*r >= 0) mpz_add_ui(tval, tval, *r);
    else mpz_sub_ui(tval, tval, -*r); //?!!
    // mpz_add_si(tval, tval, *r);

    *pk *= p;
    mpz_mod_ui(tval, tval, *pk);
    *r = mpz_get_si(tval);

    mpz_clear(tval);
}

void poly_lift_inf_root(poly_t F, int64_t invdf, int64_t p, int64_t* pk, int64_t* r) {
    mpz_t tval;
    
    mpz_init(tval); 
    poly_value(F, 1, *r, tval);
    
    mpz_mul_si(tval, tval, -invdf);
    if (*r >= 0) mpz_add_ui(tval, tval, *r);
    else mpz_sub_ui(tval, tval, -*r); //?!!
    // mpz_add_si(tval, tval, *r);

    *pk *= p;
    mpz_mod_ui(tval, tval, *pk);
    *r = mpz_get_si(tval);
    
    mpz_clear(tval);
}
